#!/usr/bin/perl
# convert iPhoto AlbumData.xml to Google Earth KML
# 2006-10-09 onohiroki@cup.com iphoto2kml

use strict;
use warnings;

use Mac::Tie::PList;
use Image::ExifTool 'ImageInfo';
use File::Glob ':glob';

my $debug = 0;

# シグネチャ
my $signature = "";

# iPhoto の初期設定の読み込み
my $iphoto_plist = Mac::Tie::PList->new_from_file(
    "$ENV{'HOME'}/Library/Preferences/com.apple.iPhoto.plist");

# iPhoto の初期設定から AlbumData.xml の位置を読み込み
my $album_data_path = bsd_glob($$iphoto_plist{'RootDirectory'}) . '/AlbumData.xml';

# iPhoto のデータベース AlbumData.xml を読み込み
my $ref               = iPhotoData($album_data_path);
my $albums            = shift @$ref;
my $master_Image_list = shift @$ref;
my $archive_Path      = shift @$ref;

# 引数の取得
my $argv          = q{};
my $new_path      = q{};
my $thumnail_path = q{};
$argv = shift @ARGV if ( $#ARGV > -1 );        # list または AlbumID
$new_path = shift @ARGV if ( $#ARGV > -1 );    # 画像のパス
   # 画像のパス指定が空の場合は，iPhoto に保存されているパスを採用
$thumnail_path = shift @ARGV
    if ( $#ARGV > -1 );    # サムネイル画像のパス

my $ans;
my $album;

# 引数の処理
# list とあったらアルバムのリストを表示
if ( $argv =~ "list" ) {
    listAlbums($albums);
    exit;
}

# 文字列がスラッシュを含んでいたらきっとパス
elsif ( $argv =~ "\/" ) {

    # 引数がいっこずれる
    $thumnail_path = $new_path;
    $new_path      = $argv;

    # 最後のアルバムを選択
    $album = $$albums[$#$albums];
}

# 文字列がスラッシュを含んでいなくて「h」を含んでいたらきっとヘルプ
elsif ( $argv =~ "h" ) {
    print
        "Usage: iphoto2kml.pl [OPTION]\n    or iphoto2kml.pl [NUMBER] [IMAGE FILE FOLDER PATH]\n";
    print "Out put a KML of iPhoto album data (the last album by default).\n";
    print "Select album by NUMBER. \n";
    print "\n";
    print "   list      List information about albums and album NUMBER.\n";
    print "   help      Display this and exit.\n";
    exit;
}

# AlbumID を得て，アルバムを選択
elsif ( $argv =~ /^[0-9]+$/ ) {
    $ans = selectAlbum( $albums, $argv );
    $album = shift @$ans;
}

# デフォルトの動作として，一番最後のアルバムを選択する
else {
    $album = $$albums[$#$albums];
}

# 選択されているアルバムを KML に変換する
print_album_kml( $album, $new_path, $thumnail_path );
exit;

# iPhoto のデータベース AlbumData.xml を読み込み
sub iPhotoData {
    my $albumData    = shift @_;
    my $plist        = Mac::Tie::PList->new_from_file($albumData);
    my $archive_Path = $$plist{'Archive Path'};
    my @data         = (
        $$plist{'List of Albums'},
        $$plist{'Master Image List'},
        $archive_Path
    );
    return \@data;
}

# 選択された iPhoto のアルバムを KML に変換して出力
sub print_album_kml {
    my $debug = 0;

    my $album         = shift @_;
    my $new_path      = shift @_;
    my $thumnail_path = shift @_;

    my $head = q{<?xml version="1.0" encoding="utf-8"?>
  <kml xmlns="http://earth.google.com/kml/2.1">};

    my $key_list      = $$album{'KeyList'};
    my $albumName     = xml_escape( $$album{'AlbumName'} );
    my $albumComments = q{};
    if ( $$album{"Comments"} ) {
        $albumComments = xml_escape( $$album{'Comments'} );
        $albumComments = '<description>'
            . $albumComments
            . $signature
            . '</description>';
    }

    $head = "$head\n <Folder>\n     <name>$albumName</name>\n";
    $head = "$head     <open>1</open>\n     $albumComments\n";

    print $head;

    #my @list_placemark = ();
    my $placemark_number = 0;
    my $placemark_number_width = 0;
    if (scalar(@$key_list) > 0) {
	$placemark_number_width = sprintf("%d", log(scalar(@$key_list))/log(10) + 1);
    }
    foreach my $image_key (@$key_list) {
        $placemark_number++;
        my $ref      = $$master_Image_list{"$image_key"};
        my $filePath = q{};
        if ( $$ref{'OriginalPath'} ) {
            $filePath = $$ref{'OriginalPath'};
        }
        else {
            $filePath = $$ref{'ImagePath'};
        }
        print STDERR "DEBUG FILEPATH $filePath \n" if ($debug);
        my $comment = q{};
        $comment = sprintf "  <description>%s</description>\n",
            xml_escape( $$ref{'Comment'} )
            if ( $$ref{'Comment'} );

        my $exifTool = new Image::ExifTool;
        $exifTool->Options(
            CoordFormat => '%.6f',
            DateFormat  => '%Y:%m:%dT%H:%M:%S'
        );
        $exifTool->ExtractInfo($filePath);
        my $image_file_path = q{};
        if ($new_path) {
            $image_file_path = $new_path . $exifTool->GetValue('FileName');
        }
        else {
            $image_file_path = $filePath;
        }
        my $date = $exifTool->GetValue('DateTimeOriginal');
        my $latR = $exifTool->GetValue('GPSLatitudeRef');
        if ($latR) {
            my $lat  = $exifTool->GetValue('GPSLatitude');
            $lat =~ s/\s+[NS]$//;
            my $lon  = $exifTool->GetValue('GPSLongitude');
            $lon =~ s/\s+[EW]$//;
            my $lonR = $exifTool->GetValue('GPSLongitudeRef');

            #print STDERR "latR $latR $lat lonR $lonR $lon\n";
            $lat = -$lat if ( $latR ne 'North' && $lat );
            $lon = -$lon if ( $lonR ne 'East'  && $lon );
            my $image_ele = q{};
            if ($thumnail_path) {
                $image_ele = sprintf
                    "<a href='%s' alt='' title='%s'><img src='%s' width='256' /></a>",
                    $image_file_path, $$ref{'Caption'}, $thumnail_path;
            }
            elsif ( $image_file_path =~ "tp\:\/\/" ) {
                $image_ele = sprintf
                    "<a href='%s' alt='' title='%s'><img src='%s' width='256' /></a>",
                    $image_file_path, $$ref{'Caption'}, $image_file_path;
            }
            else {
                $image_ele = sprintf "<img src='%s' width='256' />",
                    $image_file_path;
            }
            my $item = item_placemark(
                sprintf("%0${placemark_number_width}s", $placemark_number),
                xml_escape( $$ref{'Caption'} ),
                $date, $comment, $image_ele, $signature, $lon, $lat, );

            print $item;
        }
    }

    #print foreach (@list_placemark);
    print "</Folder>\n</kml>\n\n";
}

# Placemark 要素を生成して返す
sub item_placemark {
    my $placemark_number = shift @_;
    my $title            = shift @_;
    my $date             = shift @_;
    my $comment          = shift @_;
    my $image_ele        = shift @_;
    my $signature        = shift @_;
    my $lon              = shift @_;
    my $lat              = shift @_;

    # 改行コードの変換
    $comment =~ s#\x0D\x0A#<br />#g;
    $comment =~ s#\x0D#<br />#g;
    $comment =~ s#\x0A#<br />#g;

    # Placemark 要素の生成
    my $s = "  <Placemark>
    <name>$placemark_number $title</name>
    <description><![CDATA[
    <div align='left'>
      <b>$title</b><br />
      $date<br />
      $comment<br /><br/>
    </div>
    <div align='center'>
      $image_ele
    </div>
    $signature
    ]]></description>
    <Style>
       <IconStyle>
          <Icon>
             <href>http://maps.google.com/mapfiles/kml/pal4/icon46.png</href>
          </Icon>
       </IconStyle>
    </Style>
    <Point>
       <coordinates>$lon,$lat,0</coordinates>
    </Point>
  </Placemark>\n";
    return $s;
}

sub listAlbums {

    # iPhoto のアルバムのリストを生成して表示する
    my $ref = shift @_;
    print "   AlbumId PhotoCount AlbumName \n";
    foreach (@$ref) {
        my $AlbumId    = $$_{'AlbumId'};
        my $AlbumName  = $$_{'AlbumName'};
        my $PhotoCount = $$_{'PhotoCount'};
        printf "%10d   %6d   %s\n", $AlbumId, $PhotoCount, $AlbumName;

        #while (my ($key,$val) = each %$_) { print "$key => $val\n"; }
    }
}

sub selectAlbum {

    # AlbumId に対応するアルバムを選択する
    my $albums = shift @_;
    my $id     = shift @_;    # AlbumId
    my $found  = 0;
    my $album;
    foreach my $ref (@$albums) {
        my $test = $$ref{'AlbumId'};
        $album = $ref;
        if ( $test == $id ) {
            $found = 1;
            last;
        }
    }
    my @ans = ( $album, $found );
    return \@ans;
}

sub xml_escape {
    my $text = shift @_;

    $text =~ s/&/&amp;/g;

    $text =~ s/&amp\;nbsp\;/&#160\;/g
        ;    # no-break space = non-breaking space, U+00A0 ISOnum
    $text =~ s/&amp\;iexcl\;/&#161\;/g
        ;    # inverted exclamation mark, U+00A1 ISOnum
    $text =~ s/&amp\;cent\;/&#162\;/g;   # cent sign, U+00A2 ISOnum
    $text =~ s/&amp\;pound\;/&#163\;/g;  # pound sign, U+00A3 ISOnum
    $text =~ s/&amp\;curren\;/&#164\;/g; # currency sign, U+00A4 ISOnum
    $text =~ s/&amp\;yen\;/&#165\;/g;    # yen sign = yuan sign, U+00A5 ISOnum
    $text =~ s/&amp\;brvbar\;/&#166\;/g
        ;    # broken bar = broken vertical bar, U+00A6 ISOnum
    $text =~ s/&amp\;sect\;/&#167\;/g;    # section sign, U+00A7 ISOnum
    $text =~ s/&amp\;uml\;/&#168\;/g
        ;    # diaeresis = spacing diaeresis, U+00A8 ISOdia
    $text =~ s/&amp\;copy\;/&#169\;/g;    # copyright sign, U+00A9 ISOnum
    $text =~ s/&amp\;ordf\;/&#170\;/g
        ;    # feminine ordinal indicator, U+00AA ISOnum
    $text =~ s/&amp\;laquo\;/&#171\;/g
        ; # left-pointing double angle quotation mark = left pointing guillemet, U+00AB ISOnum
    $text =~ s/&amp\;not\;/&#172\;/g;    # not sign, U+00AC ISOnum
    $text =~ s/&amp\;shy\;/&#173\;/g
        ;    # soft hyphen = discretionary hyphen, U+00AD ISOnum
    $text =~ s/&amp\;reg\;/&#174\;/g
        ;    # registered sign = registered trade mark sign, U+00AE ISOnum
    $text =~ s/&amp\;macr\;/&#175\;/g
        ;    # macron = spacing macron = overline = APL overbar, U+00AF ISOdia
    $text =~ s/&amp\;deg\;/&#176\;/g;     # degree sign, U+00B0 ISOnum
    $text =~ s/&amp\;plusmn\;/&#177\;/g
        ;    # plus-minus sign = plus-or-minus sign, U+00B1 ISOnum
    $text =~ s/&amp\;sup2\;/&#178\;/g
        ;   # superscript two = superscript digit two = squared, U+00B2 ISOnum
    $text =~ s/&amp\;sup3\;/&#179\;/g
        ; # superscript three = superscript digit three = cubed, U+00B3 ISOnum
    $text =~ s/&amp\;acute\;/&#180\;/g
        ;    # acute accent = spacing acute, U+00B4 ISOdia
    $text =~ s/&amp\;micro\;/&#181\;/g;    # micro sign, U+00B5 ISOnum
    $text =~ s/&amp\;para\;/&#182\;/g
        ;    # pilcrow sign = paragraph sign, U+00B6 ISOnum
    $text =~ s/&amp\;middot\;/&#183\;/g
        ;    # middle dot = Georgian comma = Greek middle dot, U+00B7 ISOnum
    $text =~ s/&amp\;cedil\;/&#184\;/g
        ;    # cedilla = spacing cedilla, U+00B8 ISOdia
    $text =~ s/&amp\;sup1\;/&#185\;/g
        ;    # superscript one = superscript digit one, U+00B9 ISOnum
    $text =~ s/&amp\;ordm\;/&#186\;/g
        ;    # masculine ordinal indicator, U+00BA ISOnum
    $text =~ s/&amp\;raquo\;/&#187\;/g
        ; # right-pointing double angle quotation mark = right pointing guillemet, U+00BB ISOnum
    $text =~ s/&amp\;frac14\;/&#188\;/g
        ;  # vulgar fraction one quarter = fraction one quarter, U+00BC ISOnum
    $text =~ s/&amp\;frac12\;/&#189\;/g
        ;    # vulgar fraction one half = fraction one half, U+00BD ISOnum
    $text =~ s/&amp\;frac34\;/&#190\;/g
        ; # vulgar fraction three quarters = fraction three quarters, U+00BE ISOnum
    $text =~ s/&amp\;iquest\;/&#191\;/g
        ;    # inverted question mark = turned question mark, U+00BF ISOnum
    $text =~ s/&amp\;Agrave\;/&#192\;/g
        ; # latin capital letter A with grave = latin capital letter A grave, U+00C0 ISOlat1
    $text =~ s/&amp\;Aacute\;/&#193\;/g
        ;    # latin capital letter A with acute, U+00C1 ISOlat1
    $text =~ s/&amp\;Acirc\;/&#194\;/g
        ;    # latin capital letter A with circumflex, U+00C2 ISOlat1
    $text =~ s/&amp\;Atilde\;/&#195\;/g
        ;    # latin capital letter A with tilde, U+00C3 ISOlat1
    $text =~ s/&amp\;Auml\;/&#196\;/g
        ;    # latin capital letter A with diaeresis, U+00C4 ISOlat1
    $text =~ s/&amp\;Aring\;/&#197\;/g
        ; # latin capital letter A with ring above = latin capital letter A ring, U+00C5 ISOlat1
    $text =~ s/&amp\;AElig\;/&#198\;/g
        ; # latin capital letter AE = latin capital ligature AE, U+00C6 ISOlat1
    $text =~ s/&amp\;Ccedil\;/&#199\;/g
        ;    # latin capital letter C with cedilla, U+00C7 ISOlat1
    $text =~ s/&amp\;Egrave\;/&#200\;/g
        ;    # latin capital letter E with grave, U+00C8 ISOlat1
    $text =~ s/&amp\;Eacute\;/&#201\;/g
        ;    # latin capital letter E with acute, U+00C9 ISOlat1
    $text =~ s/&amp\;Ecirc\;/&#202\;/g
        ;    # latin capital letter E with circumflex, U+00CA ISOlat1
    $text =~ s/&amp\;Euml\;/&#203\;/g
        ;    # latin capital letter E with diaeresis, U+00CB ISOlat1
    $text =~ s/&amp\;Igrave\;/&#204\;/g
        ;    # latin capital letter I with grave, U+00CC ISOlat1
    $text =~ s/&amp\;Iacute\;/&#205\;/g
        ;    # latin capital letter I with acute, U+00CD ISOlat1
    $text =~ s/&amp\;Icirc\;/&#206\;/g
        ;    # latin capital letter I with circumflex, U+00CE ISOlat1
    $text =~ s/&amp\;Iuml\;/&#207\;/g
        ;    # latin capital letter I with diaeresis, U+00CF ISOlat1
    $text
        =~ s/&amp\;ETH\;/&#208\;/g; # latin capital letter ETH, U+00D0 ISOlat1
    $text =~ s/&amp\;Ntilde\;/&#209\;/g
        ;    # latin capital letter N with tilde, U+00D1 ISOlat1
    $text =~ s/&amp\;Ograve\;/&#210\;/g
        ;    # latin capital letter O with grave, U+00D2 ISOlat1
    $text =~ s/&amp\;Oacute\;/&#211\;/g
        ;    # latin capital letter O with acute, U+00D3 ISOlat1
    $text =~ s/&amp\;Ocirc\;/&#212\;/g
        ;    # latin capital letter O with circumflex, U+00D4 ISOlat1
    $text =~ s/&amp\;Otilde\;/&#213\;/g
        ;    # latin capital letter O with tilde, U+00D5 ISOlat1
    $text =~ s/&amp\;Ouml\;/&#214\;/g
        ;    # latin capital letter O with diaeresis, U+00D6 ISOlat1
    $text =~ s/&amp\;times\;/&#215\;/g;   # multiplication sign, U+00D7 ISOnum
    $text =~ s/&amp\;Oslash\;/&#216\;/g
        ; # latin capital letter O with stroke = latin capital letter O slash, U+00D8 ISOlat1
    $text =~ s/&amp\;Ugrave\;/&#217\;/g
        ;    # latin capital letter U with grave, U+00D9 ISOlat1
    $text =~ s/&amp\;Uacute\;/&#218\;/g
        ;    # latin capital letter U with acute, U+00DA ISOlat1
    $text =~ s/&amp\;Ucirc\;/&#219\;/g
        ;    # latin capital letter U with circumflex, U+00DB ISOlat1
    $text =~ s/&amp\;Uuml\;/&#220\;/g
        ;    # latin capital letter U with diaeresis, U+00DC ISOlat1
    $text =~ s/&amp\;Yacute\;/&#221\;/g
        ;    # latin capital letter Y with acute, U+00DD ISOlat1
    $text =~ s/&amp\;THORN\;/&#222\;/g
        ;    # latin capital letter THORN, U+00DE ISOlat1
    $text =~ s/&amp\;szlig\;/&#223\;/g
        ;    # latin small letter sharp s = ess-zed, U+00DF ISOlat1
    $text =~ s/&amp\;agrave\;/&#224\;/g
        ; # latin small letter a with grave = latin small letter a grave, U+00E0 ISOlat1
    $text =~ s/&amp\;aacute\;/&#225\;/g
        ;    # latin small letter a with acute, U+00E1 ISOlat1
    $text =~ s/&amp\;acirc\;/&#226\;/g
        ;    # latin small letter a with circumflex, U+00E2 ISOlat1
    $text =~ s/&amp\;atilde\;/&#227\;/g
        ;    # latin small letter a with tilde, U+00E3 ISOlat1
    $text =~ s/&amp\;auml\;/&#228\;/g
        ;    # latin small letter a with diaeresis, U+00E4 ISOlat1
    $text =~ s/&amp\;aring\;/&#229\;/g
        ; # latin small letter a with ring above = latin small letter a ring, U+00E5 ISOlat1
    $text =~ s/&amp\;aelig\;/&#230\;/g
        ;    # latin small letter ae = latin small ligature ae, U+00E6 ISOlat1
    $text =~ s/&amp\;ccedil\;/&#231\;/g
        ;    # latin small letter c with cedilla, U+00E7 ISOlat1
    $text =~ s/&amp\;egrave\;/&#232\;/g
        ;    # latin small letter e with grave, U+00E8 ISOlat1
    $text =~ s/&amp\;eacute\;/&#233\;/g
        ;    # latin small letter e with acute, U+00E9 ISOlat1
    $text =~ s/&amp\;ecirc\;/&#234\;/g
        ;    # latin small letter e with circumflex, U+00EA ISOlat1
    $text =~ s/&amp\;euml\;/&#235\;/g
        ;    # latin small letter e with diaeresis, U+00EB ISOlat1
    $text =~ s/&amp\;igrave\;/&#236\;/g
        ;    # latin small letter i with grave, U+00EC ISOlat1
    $text =~ s/&amp\;iacute\;/&#237\;/g
        ;    # latin small letter i with acute, U+00ED ISOlat1
    $text =~ s/&amp\;icirc\;/&#238\;/g
        ;    # latin small letter i with circumflex, U+00EE ISOlat1
    $text =~ s/&amp\;iuml\;/&#239\;/g
        ;    # latin small letter i with diaeresis, U+00EF ISOlat1
    $text =~ s/&amp\;eth\;/&#240\;/g; # latin small letter eth, U+00F0 ISOlat1
    $text =~ s/&amp\;ntilde\;/&#241\;/g
        ;    # latin small letter n with tilde, U+00F1 ISOlat1
    $text =~ s/&amp\;ograve\;/&#242\;/g
        ;    # latin small letter o with grave, U+00F2 ISOlat1
    $text =~ s/&amp\;oacute\;/&#243\;/g
        ;    # latin small letter o with acute, U+00F3 ISOlat1
    $text =~ s/&amp\;ocirc\;/&#244\;/g
        ;    # latin small letter o with circumflex, U+00F4 ISOlat1
    $text =~ s/&amp\;otilde\;/&#245\;/g
        ;    # latin small letter o with tilde, U+00F5 ISOlat1
    $text =~ s/&amp\;ouml\;/&#246\;/g
        ;    # latin small letter o with diaeresis, U+00F6 ISOlat1
    $text =~ s/&amp\;divide\;/&#247\;/g;    # division sign, U+00F7 ISOnum
    $text =~ s/&amp\;oslash\;/&#248\;/g
        ; # latin small letter o with stroke, = latin small letter o slash, U+00F8 ISOlat1
    $text =~ s/&amp\;ugrave\;/&#249\;/g
        ;    # latin small letter u with grave, U+00F9 ISOlat1
    $text =~ s/&amp\;uacute\;/&#250\;/g
        ;    # latin small letter u with acute, U+00FA ISOlat1
    $text =~ s/&amp\;ucirc\;/&#251\;/g
        ;    # latin small letter u with circumflex, U+00FB ISOlat1
    $text =~ s/&amp\;uuml\;/&#252\;/g
        ;    # latin small letter u with diaeresis, U+00FC ISOlat1
    $text =~ s/&amp\;yacute\;/&#253\;/g
        ;    # latin small letter y with acute, U+00FD ISOlat1
    $text =~ s/&amp\;thorn\;/&#254\;/g
        ;    # latin small letter thorn, U+00FE ISOlat1
    $text =~ s/&amp\;yuml\;/&#255\;/g
        ;    # latin small letter y with diaeresis, U+00FF ISOlat1

    $text =~ s/&amp\;(#\d+\;)/&$1/g;

    $text =~ s/&amp\;lt\;/&lt\;/g;
    $text =~ s/&amp\;gt\;/&gt\;/g;
    $text =~ s/&amp\;aps\;/&apos;/g;
    $text =~ s/&amp\;quot;/&quot;/g;
    $text =~ s/\'/&apos;/g;
    $text =~ s/\"/&apos;/g;
    $text =~ s/</&lt\;/g;
    $text =~ s/>/&gt\;/g;

    return $text;
}
